{-# LANGUAGE GADTs, FlexibleInstances #-}
module RandomNumbers where

import System.Random (StdGen, randomR)
import Data.List ((\\))

import Program

-- Random a  =  random variable taking values of type a

type Random a = Program RandomInstruction a

data RandomInstruction a where
    Uniform :: [a] -> RandomInstruction a

uniform :: [a] -> Random a
uniform = singleton . Uniform

-- Interpreters
    -- sampling
sample :: Random a -> StdGen -> (a,StdGen)
sample (Return a)             gen = (a,gen)
sample (Uniform xs `Then` is) gen = sample (is $ xs !! k) gen'
    where (k,gen') = System.Random.randomR (0,length xs-1) gen

    -- calculate full probability distribution
type Probability = Rational
distribution :: Random a -> [(a,Probability)]
distribution (Return a)             = [(a,1)]
distribution (Uniform xs `Then` is) = 
    [(a,p/n) | x <- xs, (a,p) <- distribution (is x)]
    where n = fromIntegral (length xs)

-- extended Interpreters
    -- infinite list of samples
samples :: Random a -> StdGen -> [a]
samples r gen = a : samples r gen'
    where (a,gen') = sample r gen

    -- calculate probability of an event
(??) :: (a -> Bool) -> Random a -> Probability
event ?? r = sum [p | (a,p) <- distribution r, event a]


-- Examples
    -- dice
die :: Random Int
die = uniform [1..6]

    -- try for instance
    --      (== 7) ?? sum2Dice
sum2Dice = die >>= \a -> die >>= \b -> return (a+b)


    -- the Monty Hall Problem
    -- http://en.wikipedia.org/wiki/Monty_Hall_problem
    -- 
    -- Returns  True  if switching was worth it. Try
    --      (== True) ?? montyHall
montyHall :: Random Bool
montyHall = let doors = [1,2,3] in do
    car    <- uniform doors                   -- door containing car
    player <- uniform doors                   -- door opened by player
    host   <- uniform $ doors \\ [car,player] -- door opened by host
    return $ [car] == doors \\ [player,host]  -- True if switching wins

